/*	Overview:
 *
 *	The assumed version of LISP has a set of blocks, each BLOCK_SIZE
 * bytes large. These blocks are where CONS-cells are to be allocated
 * by default, as the Garbage-Collection of malloc'ed objects is less
 * efficient. At the base address of each, there will be one cell
 * which will never be returned by the allocator for use by the
 * interpreter. Its CDR field will point to the base address of the
 * next block, or be NIL in the case of the last allocated block.
 *
 * 	The CAR field of the base-cell will either be NIL, or point in,
 * to a sub-list of free cells for any one block. The CDR fields of
 * cells in these sub-lists will contain tag bits, consistently with
 * how the LISP interpreter works. Also, the last free cell in
 * any block will have NIL as its CDR-value.
 *	It's irrelevant whether the CDR field of the base cells has
 * tag-bits.
 *
 * 	What's tricky about this allocator, which takes from a block
 * of CONS-cells, is the fact that each pair of blocks must be locked
 * with a Mutex, even if it can offer no free cells, but, that the
 * operation must remain locked if the returned value was not NIL.
 * 	For the sake of simplicity, this allocator will unlock the
 * pair of blocks, when it returns NIL.
 *
 */

#include "Allocator.h"

/*	Firstly, there is a set of inlined functions, which can therefore
 * not be compiled-in to any object file, but the defintion of
 * which needs to be repeated in each source file, because they've
 * been declared in LISP.h ...
 */

INLINE size_t ClearBits(size_t &datum, size_t bits) {
    size_t temp = datum | bits;
    return temp ^ bits;
}

INLINE size_t * MakeAddress1(size_t &field) {
    return (size_t *) ClearBits(field, MSB | 3);
}

INLINE CONS_cell * MakeAddress2(size_t &field) {
    return (CONS_cell *) ClearBits(field, MSB | 3);
}

INLINE bool DetermineWhite(CONS_cell &cell) {
    return (! (cell.CAR & MSB));
}

INLINE bool DetermineWhite(size_t &field) {
    //  For Immediate type Address.
    return (! (field & MSB));
}

/*	...  */

INLINE void Mark_field(size_t &field) {
    field |= MSB;
}


//	Now, we can get to defining the allocator..

INLINE CONS_cell * AllocateFrBlock(CONS_cell *base) {
    CONS_cell *retval = MakeAddress2(base->CAR);

    assert (retval);

    base->CAR = ClearBits(retval->CDR, MSB | 3);
    return retval;
}

EXPORT_INLINE CONS_cell * AllocateCONS(size_t &present, size_t &next,
                                       ChoiceMutex *m) {
    CONS_cell *temp = NULL;

    while (present) {
        next = present + BLOCK_SIZE;

        m->Reserve(present);
        m->Reserve(next);

        temp = MakeAddress2(present);
        if (temp->CAR) return AllocateFrBlock(temp);

        m->Release(present);
        m->Release(next);

        present = temp->CDR;

    }

    return NULL;

}

// This function will create a free-sub-list for one block,
// and already mark all its cells to black, if the macro is defined.
INLINE void InitNewBlock(CONS_cell *base) {
    int i = (BLOCK_SIZE / (sizeof(CONS_cell))) - 1;
    size_t last = 0;
    CONS_cell *temp = NULL;

    for ( ; i > 0 ; i-- ) {
        temp = base + i;

#ifdef PF_CALLED

        if (last) last |= MSB | 1;

        temp->CDR = last;	// If this is just NIL, it doesn't matter.
        // DetermineWhite() solely uses the CAR.
        temp->CAR = MSB;

#else

        if (last) last |= 1;
        temp->CDR = last;
        temp->CAR = 0;

#endif

        last = (size_t) temp;
    }

    base->CAR = last;
}

/*	AFAICT, the following function Should Not require any Mutexes.
 * The newly-malloc'ed block is at a memory location which the rest
 * of LISP is not aware of yet. And, As Long As only 1 thread is
 * allowed to run this function, its only feature which 'writes-in'
 * to a location that the rest of LISP is aware of, does so with
 * two instructions:
 *
 * new_base->CDR = (size_t) root;
 * root = new_base;
 *
 *	According to the GC, those instructions either have or have not
 * taken place. In theory, the current garbage-collection
 * cycle may just completely overlook the newly-added block, when
 * marking the free-lists black, and then later, because the block
 * became known, decide to reap them, thereby wasting some CPU cycles.
 *
 *	In practice, the way I implemented InitNewBlock() sidesteps
 * such waste.
 *
 *	Yet, just in case more than one thread wants to run this function,
 * I gave it a Mutex lock.
 *
 */

bool AllocNewBlock(CONS_cell *&root, ChoiceMutex *m) {
    CONS_cell *new_base = NULL;
    void * new_block_allocated = NULL;
    size_t here = (size_t) &root;	// Pointer should be stored in
    // constant location.

    new_block_allocated = malloc(BLOCK_SIZE);
    assert (new_block_allocated);	// May replace with 'return false;'

    new_base = (CONS_cell *) new_block_allocated;
    InitNewBlock(new_base);

    m->Reserve(here);
    new_base->CDR = (size_t) root;
    root = new_base;
    m->Release(here);

    return true;
}

CONS_cell * SafeAllocateCONS(size_t &present, size_t &next,
                             CONS_cell *&root, ChoiceMutex *m) {

    if (!root) AllocNewBlock(root, m);

    present = (size_t) root;
    CONS_cell *retval = AllocateCONS(present, next, m);
    if (retval) return retval;
    // No Mutexes should be locked, if execution continues here.

    AllocNewBlock(root, m);
    present = (size_t) root;
    assert (present);

    retval = AllocateCONS(present, next, m);
    assert (retval);

    return retval;
}

void RegisterMalloc(void *obj, CONS_cell *&list_root,
                    CONS_cell *&first_block, ChoiceMutex *m) {
    size_t present = 0;
    size_t next = 0;
    CONS_cell *cell = NULL;
    size_t *proxy = (size_t *) obj;

    Mark_field(*proxy);
    cell = SafeAllocateCONS(present, next, first_block, m);
    m->Reserve(0);
    // 3 Mutexes are locked now !

    cell->CDR = ((size_t) list_root) | 1;
    cell->CAR = (size_t) obj;

    /*	Hypothetically, the following code would freeze, since
     * the present thread already has the Mutexes, but,
     * the function-calls below would try to acquire them again...
    Mark_cell_gray(*cell, m);
    Mark_cell_black(*cell, m);
     */

    cell->CAR |= MSB;
    cell->CDR |= MSB;

    m->Release(present);
    m->Release(next);
    // 2 Previous Mutexes are unlocked, 1 to go.

    list_root = cell;
    m->Release(0);
    // All Mutexes should be unlocked now.
}
