#ifndef LISP_H
#define LISP_H


#include <atomic>
#include <iostream>
#include <cassert>
#include <memory>
#include <unordered_map>

#include <stdint.h>


#if _WIN32 || _WIN64
#if _WIN64
#define ENVIRONMENT64
#else
#define ENVIRONMENT32
#endif
#endif

// Check GCC
#if __GNUC__
#if __x86_64__ || __ppc64__
#define ENVIRONMENT64
#else
#define ENVIRONMENT32
#endif
#endif


// The following macros are understood, only to work with 'global
// functions', aka 'free functions'. (Not, with 'methods'.) ...

#ifndef INLINE
# if __GNUC__ && !__GNUC_STDC_INLINE__
#  define INLINE extern inline
#  define EXTERN_INLINE inline
#  define EXPORT_INLINE inline
# else
#  define INLINE inline
#  define EXTERN_INLINE extern inline
#  define EXPORT_INLINE inline
# endif
#endif


/*	A basic policy being followed is, that size_t integers will have
 * tag-bits set according to LISP.
 *	However, all pointers to CONS_cell or to size_t must not have them.
 * To that effect, the functions MakeAddress1() or MakeAddress2()
 * should be used whenever possible.
 *	Further, it's assumed that, regardless of the order of CAR and CDR
 * as part of this struct, pointers must always point to its starting
 * address. For that reason, CDR can not really be,
 * ~Complement Decrement Register~ .
 *
 */

struct CONS_cell {
    size_t CAR;
    size_t CDR;
};

#ifdef ENVIRONMENT64
#define MSB 0x8000000000000000
#else
#ifdef ENVIRONMENT32
#define MSB 0x80000000
#endif
#endif

#define BLOCK_SIZE 0x10000
#define BLOCK_POW 16


#ifdef _WIN32
#include <Windows.h>
#include <processthreadsapi.h>
#define GET_THREAD_ID GetCurrentThreadId()
#define TID_TYPE DWORD
#define THREAD_COMP(a, b) a == b
#else
#include <unistd.h>
#include <thread>
#define GET_THREAD_ID pthread_self()
#define TID_TYPE pthread_t
#define THREAD_COMP(a, b) pthread_equal(a, b)
#endif

/* The following declarations only serve as reminders,
 * of what must be redefined in every source file
 * that uses LISP's tag bits...
 *
 */

INLINE size_t ClearBits(size_t &datum, size_t bits);
INLINE size_t * MakeAddress1(size_t &field);
INLINE CONS_cell * MakeAddress2(size_t &field);

INLINE bool DetermineWhite(CONS_cell &cell);
INLINE bool DetermineWhite(size_t &field);

/*	...  */


class ChoiceMutex;	// Forward declaration.


// The functions below are actually compiled into the object files...


/*	The GC will use the following function to intialize the
 * CONS-cell blocks for tricolor mapping.
 * 'root' Must point to base of first block.
 *
 */

extern void InitListsToWhite(CONS_cell *root, ChoiceMutex *m);

/* The GC will UnMark the non-immediate data-type
 * objects, listed by 'root', that have merely been malloc'ed,
 * using the following functions.
 *
 */

extern void PreserveMapList(CONS_cell *root, ChoiceMutex *m);
extern void UnMarkCARs(CONS_cell *root);

/*	In a way that exceeds what PreserveMapList() does, there
 * also exists a needed function, from which the CAR fields of a
 * list are Not Ignored, but rather, all marked gray, just so that
 * LISP's various stacks can be marked gray efficiently...
 *
 * 	Theoretically, just marking the first cell of such a list
 * gray, should accomplish the same thing eventually. But,
 * that would only proceed at the rate of 2 cells / pass. In some
 * cases, an entire list must be marked gray quickly...
 *
 *	Note: The interpreter must mark each cell black, Before
 * splicing it in front of this list (when performing a Push).
 * After that, updating (list_root) needs to be atomic.
 *
 */

extern void Mark_list_gray(CONS_cell *list_root, ChoiceMutex *m);

/*	The following function, when run by the GC, will Mark the entire
 * free-pool black. 'root' Must point to base of the first block.
 *
 *	However, the use of this function is optional. It was conceived to
 * accelerate how fast ReapLists() will later run. However, there
 * was no way to prevent this function from acquiring locks on
 * blocks itself. So, it could slow things down instead.
 *
 *	If this function is NOT to be called, then the macro PF_CALLED
 * must be commented out below, so that the function 'ReapBlock()',
 * defined in the file 'Reaper.cpp', will compile differently,
 * as shown in that file.
 *
 */

extern void PreserveFree(CONS_cell *root, ChoiceMutex *m);
#define PF_CALLED

// The next function will Mark a specified, root, CONS-cell gray.
// Only the GC is allowed to Mark anything gray.

extern void EnterCellMapping(CONS_cell &cell, ChoiceMutex *m);

/* Here it Must be true, that the address doesn't really point
 * to a CONS-cell, but Is rather, of immediate type Address or NIL...
 * Its (CDR field) will be ignored. */

EXTERN_INLINE void Mark_address_gray(CONS_cell *address,
                                     ChoiceMutex *m);


// Next, the GC will repetitively scan all the blocks for gray cells
// until none are left, and all the white cells are unreachable.

extern void MapBlockList(CONS_cell *first_block, ChoiceMutex *m);


/*	After mapping comes 'reaping'. This is also, the 'recyling' of
 * relevant cells. Here, the GC passes-in the base cell of the first
 * CONS block as 'root'.
 *
 */

extern long int ReapLists(CONS_cell *root, ChoiceMutex *m);


/*	In order to reap all malloc'ed objects (of Immediate type "Address")
 * which have Not been marked, the GC calls the following function.
 * Beware that here, 'root' is both input and output.
 *
 *	Hence, a reference to a pointer is being passed in, to the list
 * that keeps track of all the objects, and the function must be
 * allowed to change that pointer itself.
 *
 *	Because of that, both the GC and the Interpreter must
 * competitively lock the Mutex on resource (0),
 *
 *	(the Interpreter) After allocating, but before editing this List,
 *	(the GC) Inside the call of this function.
 *
 * 	The Interpreter must also Mark the CONS-cells black, after
 * allocating them, and before releasing their respective Mutex locks.
 *
 *	This function depends on MapBlockList(), to detect which objects
 * in the list are still linked to.
 *
 */

extern int FreeNonImmediate(CONS_cell *&root, ChoiceMutex *m);



/*	These next functions should be used by the Interpreter thread
 * to allocate a CONS-cell, and to stay within my use of Mutexes.
 * They will return with the Mutexes on 'present' and 'next' locked.
 * Those are Both outputs!
 *
 *	But, when invoking AllocateCONS(), 'present' must also be set to
 * The Base Of the first block, as an input, when the function is
 * first invoked.
 *
 *	Further, for convenience, this first function will unlock 'present'
 * and 'next', if NULL is returned. This will only happen if:
 *
 *	1: There are no more free cells in all the blocks, Or
 *	2: The LISP state's memory is not set up correctly (Oops!)
 *
 *	Note: Consistently with GC policy, the Mutexes should only be
 * unlocked, after a returned cell has been Marked 'black', by the
 * Interpreter.
 *
 */

EXTERN_INLINE CONS_cell * AllocateCONS(size_t &present, size_t &next,
                                       ChoiceMutex *m);

/*	Because it may seem inconvenient, if NIL was returned
 * by the previous function, and a new block must be allocated,
 * the following convenience function is provided, which will
 * do so automatically.
 *
 *	Note: In the following function, 'root' is both input and
 * output. If a new block had to be allocated, then this function
 * must also be able to change the pointer which the system keeps,
 * to the logical first block.
 *
 *	'present' and 'next' will purely be output, as resources that have
 * a Mutex lock, after the function-call.
 *
 */

extern CONS_cell * SafeAllocateCONS(size_t &present, size_t &next,
                                    CONS_cell *&root, ChoiceMutex *m);

/*	Some LISP interpreters may not want to wait, until the CONS pool
 * runs out, before allocating new blocks. In this case, new blocks
 * can just as easily be added at any time, to keep a reserve, using
 * the following function. Again, 'root' is both input and output.
 *
 *	The present version of this function will either return true
 * or fail an assertion that a new block could be malloc'ed .
 * Future versions may return false in the case of such a failure.
 *
 */

extern bool AllocNewBlock(CONS_cell *&root, ChoiceMutex *m);

/*	Another premise of the LISP implementation is, that the Interpreter
 * will casually use malloc() to allocate objects, the immediate
 * data-type of which will be 'Address', but the non-immediate
 * data-type of which is indicated by their first byte. Also,
 * this first-byte value must be less than 128, because if it is not,
 * the objects would be allocated in the CONS blocks, and their
 * only other, second word would be a list.
 *
 *	LISP is to keep a List, which itself exists in the CONS blocks,
 * where all such malloc'ed objects are registered.
 *
 *	The following function will register an object via (void *).
 * 'list_root' and 'first_block_base' Must be usable as both
 * input and output.
 *
 */

extern void RegisterMalloc(void *obj, CONS_cell *&list_root,
                           CONS_cell *&first_block_base, ChoiceMutex *m);


#endif
