#ifndef REAPER_H
#define REAPER_H

#include "LISP.h"

/*	This header file must not only be included by the reaper, but also
 * by any source file that is to share the Mutex with it...
 */


// The following is actually compiled into the object file...


// On the assumption that no more CONS-cells are gray, this function
// is called to reap / recycle all the cells marked white.

extern long int ReapLists(CONS_cell *root, ChoiceMutex *m);


/*	This LISP posesses a List of objects not in the CONS blocks,
 * that were simply malloc'ed, and have unknown word-counts.
 * On the assumption that all non-immediate data which are linked
 * have been Marked, the following function will free
 * the ones which are still UnMarked... */

extern void FreeNonImmediate(CONS_cell *&root);



class Mutex {
private:
    std::atomic<int> state;
    //  One could pretend that this data-type exists
    //  natively.

    typedef std::atomic<int> *SharedAtomicIntPtr;
    //  Any void pointer returned from a shared memory segment
    //  will need to be cast to this pointer-type, then fed to
    //  'atomic_store_explicit()' and 'atomic_exchange_explicit()' below.

    SharedAtomicIntPtr sharedState;

public:
    Mutex() {
        sharedState = &state;
        std::atomic_store_explicit(sharedState,
                                   0, std::memory_order_release);
    }

    explicit Mutex(void *sharedPointer) {
        sharedState = (SharedAtomicIntPtr) sharedPointer;
        std::atomic_store_explicit(sharedState,
                                   0, std::memory_order_release);
    }

    void Unlock() {
        std::atomic_store_explicit(sharedState,
                                   0, std::memory_order_release);
    }

    void Lock(double laxity=20.0) {
        if (laxity > 0.0) {
            while (std::atomic_exchange_explicit(sharedState,
                                                 1, std::memory_order_acquire)) {
                sleep(laxity / 1000.0);
            }
        } else {
            while (std::atomic_exchange_explicit(sharedState,
                                                 1, std::memory_order_acquire)) {
            }
        }
    }

    bool IsLocked() {
        return (*sharedState != 0);
    }
};

class ChoiceMutex : public Mutex {
private:

    // Only the constructor call sets this.
    size_t m_mask;	// High-order bits.

    std::unordered_map<size_t, uint32_t> resource_served_m;
    std::unordered_map<size_t, uint32_t> resource_asked_m;
    /* The unsigned integers just wrap around
    when they overflow. Thus, one doesn't
    need to worry about the range of
    numbers 'running out'. */
public:
    explicit ChoiceMutex(unsigned int mutex_pow) : Mutex() {
        size_t mask = 1;

        assert (mutex_pow < sizeof(size_t) * 8);

        if (mutex_pow > 0) mask = mask << mutex_pow;
        m_mask = ~( --mask ) ;
    }

    void Reserve(size_t choice) {
        uint32_t cue = 0;

        choice = choice & m_mask;

        Mutex::Lock(1.0);

        if (resource_asked_m.count(choice)) {
            resource_asked_m[choice] += 1;
            cue = resource_asked_m[choice] - 1;
        } else {
            resource_asked_m[choice] = 1;
        }

        if (! resource_served_m.count(choice))
            resource_served_m[choice] = 0;

        Mutex::Unlock();

        while (resource_served_m[choice] != cue) {
            // It's important not to use
            // an inequality here, as
            // both values wrap around !
            sleep(0.001);
        }
    }

    void Release(size_t choice) {

        assert (resource_served_m.count(choice) != 0);

        Mutex::Lock(1.0);
        resource_served_m[choice] += 1;
        Mutex::Unlock();

    }
};


#endif
